@@ -516,7 +516,7 @@ void QPCLS(matrix *Z,matrix *X, matrix *p, matrix *y,matrix *Ain,matrix *b,matri
 
 
 void PCLS(matrix *X,matrix *p,matrix *y,matrix *w,matrix *Ain,matrix *b,
-          matrix *Af,matrix *H,matrix *S,int *off,double *theta,int m,int *active)
+          matrix *Af,matrix *S,int *off,double *theta,int m,int *active)
 
 /* Routine for Penalized Constrained Least Squares problems.
    PCLS() is an interface routine for QPCLS for solving the general problem class:
@@ -538,9 +538,6 @@ void PCLS(matrix *X,matrix *p,matrix *y,matrix *w,matrix *Ain,matrix *b,
    ... where F = [ X'W^0.5, B^0.5']'  and z = [y'W^0.5, 0]'. This rewrite is
    performed and then QPCLS is called to obtain the solution.
 
-   If H->r==y->r on entry, then an influence (or "hat") matrix is returned in H.
-   At present the calculation of H is inefficient and none too stable.
-
    On exit active[] contains a list of the active inequlity constraints in elements 
    1->active[0]. This array should be initialized to length p.r+1 on entry.
 
@@ -548,9 +545,9 @@ void PCLS(matrix *X,matrix *p,matrix *y,matrix *w,matrix *Ain,matrix *b,
 
 */
 
-{ int i,j,k;
-  matrix z,F,W,Z,B,C;
-  double x,xx;
+{ int i,j,k,n;
+  matrix z,F,W,Z,B;
+  double x,xx,*p1,*C;
  
   /* form transformed data vector z */
   if (m>0) z=initmat(y->r+p->r,1);else z=initmat(y->r,1);
@@ -562,27 +559,33 @@ void PCLS(matrix *X,matrix *p,matrix *y,matrix *w,matrix *Ain,matrix *b,
   for (i=0;i<X->r;i++) for (j=0;j<X->c;j++) F.M[i][j]=W.V[i]*X->M[i][j];
   /* add up the Penalties */
  
-  if (m>0)
-  { B=initmat(p->r,p->r);
+  if (m>0) { //B=initmat(p->r,p->r);
+    n = p->r;
+    C = (double *)CALLOC((size_t)(n*n),sizeof(double));
     for (k=0;k<m;k++) for (i=0;i<S[k].r;i++) for (j=0;j<S[k].c;j++)
-    B.M[i+off[k]][j+off[k]]+=theta[k]*S[k].M[i][j];
+					       C[i+off[k]+n*(j+off[k])] += theta[k]*S[k].M[i][j];
+    //B.M[i+off[k]][j+off[k]]+=theta[k]*S[k].M[i][j];
     /* and find a square root of B..... */
 
-    root(&B,&C,8*DOUBLE_EPS);
-
+    //root(&B,&C,8*DOUBLE_EPS); // CC' = S_tot (min cols)
+    k = -1;
+    mroot(C,&k,&n); // C'C = S_tot (min rows - returned in k)
+    /* copy C into the last p->r rows of F */
+    for (p1=C,i=0;i<n;i++) for (j=0;j<k;j++,p1++) F.M[j+X->r][i] = *p1;
     /* copy C' into the last p->r rows of F */
-    for (i=0;i<C.r;i++) for (j=0;j<C.c;j++) F.M[j+X->r][i]=C.M[i][j];
-    freemat(B);freemat(C);
+    //for (i=0;i<C.r;i++) for (j=0;j<C.c;j++) F.M[j+X->r][i]=C.M[i][j];
+    FREE(C);
+    //freemat(B);//freemat(C);
   }
   /*  printf("\ncond(F)=%g",condition(F));*/
   /* Which means that the problem is now in a form where QPCLS can solve it.... */
   QPCLS(&Z,&F,p,&z,Ain,b,Af,active); /* note that at present Z is full not HH */
-  if (H->r==y->r) /* then calculate the influence matrix XZ(Z'F'FZ)^{-1}Z'X'W */
-  { freemat(W);W=initmat(Z.c,Z.c);
-    multi(4,W,Z,F,F,Z,1,1,0,0);invert(&W); /* Wildly inefficient!! */
-    multi(5,*H,*X,Z,W,Z,*X,0,0,0,1,1);      /* ditto */
-    for (i=0;i<H->r;i++) for (j=0;j<H->c;j++) H->M[i][j]*=w->V[j];
-  }
+  //  if (H->r==y->r) /* then calculate the influence matrix XZ(Z'F'FZ)^{-1}Z'X'W */
+  //{ freemat(W);W=initmat(Z.c,Z.c);
+  //  multi(4,W,Z,F,F,Z,1,1,0,0);invert(&W); /* Wildly inefficient!! */
+  //  multi(5,*H,*X,Z,W,Z,*X,0,0,0,1,1);      /* ditto */
+  //  for (i=0;i<H->r;i++) for (j=0;j<H->c;j++) H->M[i][j]*=w->V[j];
+  //}
   /* working out value of objective at minimum */
   B=initmat(z.r,1);matmult(B,F,*p,0,0);
   xx=0.0;for (i=0;i<z.r;i++) { x=B.V[i]-z.V[i];xx+=x*x;}
